package org.tinyjee.maven.dim.extensions;

import org.apache.batik.transcoder.image.ImageTranscoder;
import org.apache.batik.transcoder.image.PNGTranscoder;
import org.tinyjee.maven.dim.utils.AbstractAliasHandler;

import java.io.File;
import java.util.HashMap;
import java.util.Map;

/**
 * SvgLoader is a 'source-class' compatible helper class that can load and transform SVG content and include it either
 * directly or rendered as PNG.
 * <p/>
 * The loader is an extension of XmlLoader and supports remote URLs (e.g. services) and local paths to get the SVG source content.
 * <p/>
 * <b>Usage:</b><div><code>
 * %{include|source=template.vm|source-class=org.tinyjee.maven.dim.extensions.SvgLoader|svg=file.svg}
 * </code></div>
 * or when using the alias:<div><code>
 * %{include|source=template.vm|source-svg=file.svg}
 * </code></div>
 * include SVG after XSL transformation:<div><code>
 * %{include|source-svg=file.xml|xsl=transform-to-svg.xsl}
 * </code></div>
 *
 * @author Juergen_Kellerer, 2011-10-06
 */
public class SvgLoader extends HashMap<String, Object> {

	private static final long serialVersionUID = -6282606147512841025L;

	/**
	 * Implements the "{@link org.tinyjee.maven.dim.extensions.SvgLoader#PARAM_ALIAS}" alias functionality.
	 */
	public static class AliasHandler extends AbstractAliasHandler {
		/**
		 * Constructs the handler (Note: Is called by {@link org.tinyjee.maven.dim.spi.RequestParameterTransformer#TRANSFORMERS}).
		 */
		public AliasHandler() {
			super(PARAM_ALIAS, PARAM_SVG, SvgLoader.class.getName());
		}
	}

	/**
	 * Defines a shortcut for the request properties 'source-class' and 'svg'.
	 * <p/>
	 * If this parameter is present inside the request parameters list, the system will effectively behave as if
	 * 'source-class' and 'svg' where set separately.
	 */
	public static final String PARAM_ALIAS = "source-svg";

	/**
	 * Sets the URL or local path to the SVG file to load inside the request parameters.
	 * <p/>
	 * The specified file is resolved in exactly the same way as when using the parameter "<code>source</code>"
	 * (see Macro reference).
	 * <p/>
	 * <i>Note:</i> When no further source is specified, the svg loader uses a built-in template to add the SVG to
	 * the output as image using Apache Batik to render it.
	 */
	public static final String PARAM_SVG = "svg";

	/**
	 * <b>Optional Parameter</b>, transform the given input (e.g. XML) to an SVG file using an XML stylesheet.
	 * <p/>
	 * See {@link XmlLoader} for details.
	 */
	public static final String PARAM_XSL = "xsl";

	/**
	 * <b>Optional Parameter</b>, sets the width in pixel.
	 */
	public static final String PARAM_WIDTH = "width";

	/**
	 * <b>Optional Parameter</b>, sets the height in pixel.
	 */
	public static final String PARAM_HEIGHT = "height";

	/**
	 * Specifies whether the SVG is rendered to an image and included.
	 * <p/>
	 * Possible values:<ul>
	 * <li>'true' or 'png': Renders a 32bit PNG image (including alpha channel).</li>
	 * <li>'jpg': Renders a JPG image.</li>
	 * <li>'false': Does not render an image.</li>
	 * </ul>
	 */
	public static final String PARAM_RENDER = "render";

	/**
	 * Specifies a background color to use when rendering an SVG to an image.
	 */
	public static final String PARAM_RENDER_BACKGROUND = "render-background";

	/**
	 * Specifies the CSS media when rendering an SVG containing CSS stylesheets with alternating media styles.
	 * <p/>
	 * E.g. "print", "screen"
	 */
	public static final String PARAM_RENDER_MEDIA = "render-media";

	/**
	 * Specifies whether the SVG is attached to the page (instead of copied straight to the document).
	 * <p/>
	 * Possible values:<ul>
	 * <li>'true': Attaches the SVG using the original filename.</li>
	 * <li>'[filename]': Attaches the SVG using given filename.</li>
	 * <li>'false': Does not attach the SVG.</li>
	 * </ul>
	 * <p/>
	 * <i>Note:</i> When an SVG is attached, it cannot be further processed by velocity, thus any velocity
	 * expressions inside the included SVG are only processed with inline SVG graphics and the addition of the macro
	 * parameter "{@code source-is-template=true}".
	 */
	public static final String PARAM_ATTACH = "attach";

	/**
	 * Is set with the parsed document object ({@link org.w3c.dom.Document}) and can be accessed from velocity templates.
	 */
	public static final String OUT_PARAM_DOCUMENT = "document";

	/**
	 * Is set with an XPath evaluator, see {@link XmlLoader} for details.
	 */
	public static final String OUT_PARAM_XPATH = "xpath";


	private static final String[][] parameterTranslations = {
			{PARAM_SVG, XmlLoader.PARAM_XML},
			{PARAM_XSL, XmlLoader.PARAM_XSL},
			{XmlLoader.OUT_PARAM_DOCUMENT, OUT_PARAM_DOCUMENT},
			{XmlLoader.OUT_PARAM_XPATH, OUT_PARAM_XPATH},
	};

	public SvgLoader(File baseDir, Map<String, Object> requestParams) throws Exception {
		applyParameterTranslations(requestParams);
		XmlLoader xmlLoader = new XmlLoader(baseDir, requestParams);
		putAll(xmlLoader);
		applyParameterTranslations(this);

		ImageTranscoder transcoder = new PNGTranscoder();
	}

	private static void applyParameterTranslations(Map<String, Object> requestParams) {
		for (String[] parameterTranslation : parameterTranslations) {
			if (requestParams.containsKey(parameterTranslation[0]))
				requestParams.put(parameterTranslation[1], requestParams.get(parameterTranslation[0]));
		}
	}
}
