package org.tinyjee.maven.dim.sources;

import org.tinyjee.maven.dim.spi.AbstractSource;

import java.io.*;
import java.net.URL;
import java.util.Map;

/**
 * Base class for source implementations that buffer (generated) content in memory.
 *
 * @author Juergen_Kellerer, 2011-10-19
 */
public abstract class AbstractBufferedSource extends AbstractSource {

	protected Content content;

	protected AbstractBufferedSource(URL sourceUrl, Map<String, Object> parameters) {
		super(sourceUrl, parameters);
	}

	/**
	 * Is called when the source content has to be created.
	 *
	 * @param writer the writer to write the source content to.
	 * @throws IOException in case of writing failed.
	 */
	protected abstract void writeSourceContent(Writer writer) throws IOException;

	@Override
	public synchronized Content getContent() {
		if (content == null) {
			String bufferedContent = null;
			IOException exception = null;
			try {
				final StringWriter writer = new StringWriter(2048);
				try {
					writeSourceContent(writer);
				} finally {
					writer.close();
				}
				bufferedContent = writer.toString();
			} catch (IOException e) {
				exception = e;
			}

			content = new BufferedSourceContent(bufferedContent, exception);
		}

		return content;
	}

	public synchronized void reset() {
		content = null;
	}

	protected class BufferedSourceContent extends SourceContent {

		protected final IOException exception;
		protected final String bufferedContent;

		protected BufferedSourceContent(String bufferedContent, IOException exception) {
			this.bufferedContent = bufferedContent;
			this.exception = exception;
		}

		@Override
		public Reader openReader() throws IOException {
			if (exception != null) throw exception;
			return new StringReader(bufferedContent);
		}

		@Override
		public String toString() {
			return "BufferedSourceContent{" +
					"sourceUrl=" + sourceUrl +
					", parameters=" + parameters +
					", exception=" + exception +
					", bufferedContent='" + bufferedContent + '\'' +
					'}';
		}
	}
}
