/*
 * Copyright 2010 - org.tinyjee.maven
 *
 *    Licensed under the Apache License, Version 2.0 (the "License");
 *    you may not use this file except in compliance with the License.
 *    You may obtain a copy of the License at
 *
 *        http://www.apache.org/licenses/LICENSE-2.0
 *
 *    Unless required by applicable law or agreed to in writing, software
 *    distributed under the License is distributed on an "AS IS" BASIS,
 *    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *    See the License for the specific language governing permissions and
 *    limitations under the License.
 */

package org.tinyjee.maven.dim.sources;

import org.apache.commons.collections.ExtendedProperties;
import org.apache.maven.doxia.logging.Log;
import org.apache.velocity.Template;
import org.apache.velocity.VelocityContext;
import org.apache.velocity.app.VelocityEngine;
import org.apache.velocity.context.Context;
import org.apache.velocity.exception.ResourceNotFoundException;
import org.apache.velocity.runtime.RuntimeServices;
import org.apache.velocity.runtime.log.LogChute;
import org.apache.velocity.runtime.resource.Resource;
import org.apache.velocity.runtime.resource.loader.ResourceLoader;
import org.apache.velocity.tools.ToolContext;
import org.apache.velocity.tools.ToolManager;
import org.apache.velocity.tools.config.EasyFactoryConfiguration;
import org.codehaus.plexus.util.FileUtils;
import org.tinyjee.maven.dim.spi.Globals;
import org.tinyjee.maven.dim.spi.Source;
import org.tinyjee.maven.dim.spi.UrlFetcher;
import org.tinyjee.maven.dim.spi.VelocityConfigurator;

import java.io.*;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.HashMap;

import static org.apache.velocity.runtime.RuntimeConstants.RUNTIME_LOG_LOGSYSTEM;
import static org.apache.velocity.runtime.RuntimeConstants.VM_LIBRARY_DEFAULT;
import static org.codehaus.plexus.util.FileUtils.basename;
import static org.codehaus.plexus.util.IOUtil.copy;
import static org.tinyjee.maven.dim.spi.VelocityConfigurator.CONFIGURATORS;

/**
 * Wraps a source content that contains a velocity template and executes it on using the Source interfaces.
 *
 * @author Juergen_Kellerer, 2010-09-04
 * @version 1.0
 */
public class TemplateSource extends AbstractBufferedSource {

	static final String TEMPLATE_ENCODING = "UTF-8";
	static final String TEMPLATE_PARAM_SOURCE_URL = "sourceUrl";

	private static final ThreadLocal<VelocityEngine> engineEnvironment = new ThreadLocal<VelocityEngine>() {
		@Override
		protected VelocityEngine initialValue() {
			try {
				final VelocityEngine ve = new VelocityEngine();
				ve.setProperty(RUNTIME_LOG_LOGSYSTEM, new TemplateLogger());
				for (VelocityConfigurator configurator : CONFIGURATORS) configurator.configureEngine(ve);
				ve.init();

				return ve;
			} catch (Exception e) {
				throw new RuntimeException(e);
			}
		}
	};

	static final ToolManager VELOCITY_TOOL_MANAGER;

	static {
		EasyFactoryConfiguration config = new EasyFactoryConfiguration();
		for (VelocityConfigurator configurator : CONFIGURATORS) configurator.configureTools(config);
		config.autoLoad();

		ToolManager toolManager = new ToolManager(false, false);
		toolManager.configure(config);
		toolManager.setUserCanOverwriteTools(true);
		VELOCITY_TOOL_MANAGER = toolManager;
	}

	private static void deleteTemporaryFile(File file) {
		if (file != null && !file.delete()) {
			Globals.getLog().warn("Failed to delete temporary file: " + file + ", trying to delete on exit.");
			file.deleteOnExit();
		}
	}

	final Source source;
	final VelocityEngine engine = engineEnvironment.get();

	Template template;

	/**
	 * Constructs a new TemplateSource.
	 *
	 * @param sourceUrl the original source URL (used for naming purposes).
	 * @param source    the source containing the template source and context params.
	 */
	public TemplateSource(URL sourceUrl, Source source) {
		super(sourceUrl, null);
		this.source = source;
	}

	@Override
	protected void writeSourceContent(Writer writer) throws IOException {
		final Content sourceContent = source.getContent();
		parameters.putAll(sourceContent.getParameters());
		if (template == null) createTemplate(sourceContent);
		renderContent(writer);
	}

	private void renderContent(Writer writer) throws IOException {
		final ToolContext toolContext = VELOCITY_TOOL_MANAGER.createContext();
		if (sourceUrl != null) toolContext.put(TEMPLATE_PARAM_SOURCE_URL, sourceUrl);

		final Context context = new VelocityContext(parameters, toolContext);
		for (VelocityConfigurator configurator : CONFIGURATORS)
			configurator.configureContext(new HashMap<String, Object>(parameters), context);

		template.merge(context, writer);
	}

	private void createTemplate(Content sourceContent) {
		final UrlSource urlSource = source instanceof UrlSource ? (UrlSource) source : null;
		File templateFile = null;
		try {
			if (urlSource != null) {
				// Static template
				final String charset = UrlFetcher.getSourceCharset(urlSource.getUrl()).name();
				template = engine.getTemplate(urlSource.getUrl().toExternalForm(), charset);
			} else {
				// Dynamically built template...
				File temporaryFolder = sourceUrl != null && "file".equalsIgnoreCase(sourceUrl.getProtocol()) ?
						new File(sourceUrl.toURI()).getParentFile() : null;

				String temporaryNamePrefix = sourceUrl == null ?
						"inline-content-" : basename(new File(sourceUrl.getPath()).getName()) + '-';
				templateFile = File.createTempFile(temporaryNamePrefix, ".vm", temporaryFolder);
				OutputStreamWriter output = new OutputStreamWriter(new FileOutputStream(templateFile), TEMPLATE_ENCODING);
				try {
					copy(sourceContent.openReader(), output);
				} finally {
					output.close();
				}

				template = engine.getTemplate(templateFile.toURI().toASCIIString(), TEMPLATE_ENCODING);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		} finally {
			deleteTemporaryFile(templateFile);
		}
	}

	/**
	 * Resolves templates from Urls.
	 */
	public static class UrlResourceLoader extends ResourceLoader {
		@Override
		public void init(ExtendedProperties extendedProperties) {
		}

		@Override
		public InputStream getResourceStream(String sourceUrl) throws ResourceNotFoundException {
			try {
				return UrlFetcher.getSource(new URL(sourceUrl), false);
			} catch (Exception ignored) {
				try {
					return UrlFetcher.getSource(Globals.getInstance().resolvePath(sourceUrl), false);
				} catch (Exception e) {
					if (VM_LIBRARY_DEFAULT.equals(sourceUrl)) {
						Globals.getLog().debug("A global default VM library '" + sourceUrl + "' was not found. Will run without it.");
						return new ByteArrayInputStream(new byte[0]);
					}
					throw new ResourceNotFoundException(sourceUrl, e);
				}
			}
		}

		@Override
		public boolean isSourceModified(Resource resource) {
			try {
				return UrlFetcher.isModified(new URL(resource.getName()));
			} catch (MalformedURLException ignored) {
				return false;
			}
		}

		@Override
		public long getLastModified(Resource resource) {
			try {
				return UrlFetcher.getLastModified(new URL(resource.getName()));
			} catch (MalformedURLException ignored) {
				return 0;
			}
		}
	}

	/**
	 * Translates velocity logs to Maven/Doxia logs.
	 */
	private static class TemplateLogger implements LogChute {

		public void init(RuntimeServices runtimeServices) throws Exception {
		}

		public void log(int i, String s) {
			log(i, s, null);
		}

		public void log(int i, String s, Throwable throwable) {
			Log log = Globals.getLog();
			switch (i) {
				case TRACE_ID:
				case DEBUG_ID:
					if (throwable == null) log.debug(s);
					else log.debug(s, throwable);
					break;
				case INFO_ID:
					// translating INFO without exception to DEBUG as Velocity is a bit to chatty and INFO-log is enabled by default.
					if (throwable == null) log.debug(s);
					else log.info(s, throwable);
					break;
				case WARN_ID:
					if (throwable == null) log.warn(s);
					else log.warn(s, throwable);
					break;
				case ERROR_ID:
					if (throwable == null) log.error(s);
					else log.error(s, throwable);
					break;
			}
		}

		public boolean isLevelEnabled(int i) {
			Log log = Globals.getLog();
			switch (i) {
				case TRACE_ID:
				case DEBUG_ID:
					return log.isDebugEnabled();
				case INFO_ID:
					return log.isInfoEnabled();
			}
			return true;
		}
	}
}
