package org.tinyjee.maven.dim.spi;

import org.apache.maven.doxia.logging.Log;
import org.apache.maven.doxia.logging.SystemStreamLog;
import org.apache.maven.doxia.macro.MacroRequest;
import org.apache.maven.doxia.sink.Sink;
import org.codehaus.plexus.util.IOUtil;
import org.codehaus.plexus.util.StringUtils;

import java.io.*;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetEncoder;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.regex.Matcher;
import java.util.zip.GZIPOutputStream;

import static java.util.Collections.synchronizedMap;
import static java.util.regex.Pattern.compile;

/**
 * Is a global, thread-local, recursive configuration singleton that may be used to exchange global settings between classes.
 *
 * @author Juergen_Kellerer, 2011-10-13
 */
public class Globals {

	static final String LB = System.getProperty("line.separator", "\n");

	private static final Map<Sink, Set<String>> attachedResources = synchronizedMap(new WeakHashMap<Sink, Set<String>>());
	private static final Map<Sink, AtomicInteger> localIdSequences = synchronizedMap(new WeakHashMap<Sink, AtomicInteger>());
	private static final AtomicInteger idSequence = new AtomicInteger();

	private static final ThreadLocal<Stack<Globals>> threadLocal = new ThreadLocal<Stack<Globals>>() {
		@Override
		protected Stack<Globals> initialValue() {
			return new Stack<Globals>();
		}
	};

	public static Globals initializeGlobals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		Globals globals = new Globals(charset, sink, request, logger, attachedIncludesPath);
		threadLocal.get().push(globals);
		return globals;
	}

	public static Globals reInitializeGlobals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		threadLocal.get().pop();
		return initializeGlobals(charset, sink, request, logger, attachedIncludesPath);
	}

	public static void cleanupGlobals() {
		Stack<Globals> globalsStack = threadLocal.get();
		globalsStack.pop();
		if (globalsStack.empty()) threadLocal.remove();
	}

	/**
	 * Returns a thread local, recursion safe instance.
	 *
	 * @return a thread local, recursion safe instance.
	 */
	public static Globals getInstance() {
		Stack<Globals> globalsStack = threadLocal.get();
		return globalsStack.empty() ? new Globals(null, null, null, null, null) : globalsStack.peek();
	}

	/**
	 * Returns a safe logger.
	 *
	 * @return a safe logger.
	 */
	public static Log getLog() {
		return getInstance().getLogger();
	}

	/**
	 * Returns the relative path (from web root) to the directory hosting the attached includes.
	 *
	 * @return the relative path (from web root) to the directory hosting the attached includes.
	 */
	public static String getAttachedIncludesDirectory() {
		return System.getProperty("org.tinyjee.maven.dim.attachedIncludesDirectory", "attached-includes").replace('\\', '/');
	}

	private Charset charset = Charset.defaultCharset();
	private boolean charsetIsUserSupplied;
	private final MacroRequest request;
	private final File attachedIncludesPath;
	private final Sink sink;
	private Log logger;

	private Globals(String charset, Sink sink, MacroRequest request, Log logger, File attachedIncludesPath) {
		this.sink = sink;
		this.request = request;
		this.logger = logger;
		this.attachedIncludesPath = attachedIncludesPath;

		if (charsetIsUserSupplied = charset != null) {
			this.charset = Charset.forName(charset);
		} else {
			String inputEncoding = System.getProperty("org.tinyjee.maven.dim.project.inputEncoding");
			if (!StringUtils.isEmpty(inputEncoding))
				this.charset = Charset.forName(inputEncoding);
		}
	}

	/**
	 * Returns the charset to use when reading text.
	 *
	 * @return the charset to use when reading text.
	 */
	public Charset getCharset() {
		return charset;
	}

	/**
	 * Returns true if the user supplied a charset with the macro call.
	 *
	 * @return true if the user supplied a charset with the macro call.
	 */
	public boolean getCharsetIsUserSupplied() {
		return charsetIsUserSupplied;
	}

	/**
	 * Returns a macro logger to use for logging to the build log.
	 *
	 * @return a macro logger to use for logging to the build log.
	 */
	public Log getLogger() {
		if (logger == null) logger = new SystemStreamLog();
		return logger;
	}

	/**
	 * Returns the next value of the monotonically increasing integer ID (IDs are consistent only for the lifetime of a
	 * single module's site build).
	 * <p/>
	 * This method is primarily meant to assign unique ids to page elements (e.g. for javascript, svg, etc).
	 *
	 * @return an increasing integer ID.
	 */
	public int nextId() {
		return idSequence.incrementAndGet();
	}

	/**
	 * Returns the next page local value of the increasing integer ID (local IDs are consistent only within a single page).
	 * <p/>
	 * This method is primarily meant to assign unique ids to page elements (e.g. for javascript, svg, etc).
	 *
	 * @return an increasing integer ID.
	 */
	public int nextLocalId() {
		AtomicInteger idSequence;
		synchronized (localIdSequences) {
			idSequence = localIdSequences.get(sink);
			if (idSequence == null) localIdSequences.put(sink, idSequence = new AtomicInteger());
		}
		return idSequence.incrementAndGet();
	}

	/**
	 * The current sink, may be 'null'.
	 *
	 * @return the current sink, may be 'null'.
	 */
	public Sink getSink() {
		return sink;
	}

	/**
	 * Creates a globals instance for the given sink.
	 * <p/>
	 * If sink is 'null' or the same as the current, the method returns 'this'.
	 *
	 * @param sink the sink to return a globals instance for.
	 * @return a globals instance that operates on the given sink.
	 */
	public Globals operateOnSink(Sink sink) {
		if (sink == null || sink == this.sink)
			return this;

		final Globals globals = new Globals(null, sink, request, logger, attachedIncludesPath);
		globals.charset = charset;
		globals.charsetIsUserSupplied = charsetIsUserSupplied;
		return globals;
	}

	/**
	 * Returns true if the sink produces some sort of HTML.
	 *
	 * @return true if the sink produces some sort of HTML.
	 */
	public boolean isHTMLCapableSink() {
		try {
			return sink instanceof org.apache.maven.doxia.markup.HtmlMarkup;
		} catch (NoClassDefFoundError ignored) {
			// Old API, use String-Search as fallback.
			String[] names = {"XhtmlSink", "XhtmlBaseSink"};
			for (Class<?> cls = sink.getClass(); cls != null; cls = cls.getSuperclass())
				for (String name : names)
					if (cls.getName().contains(name))
						return true;
			return false;
		}
	}

	/**
	 * The current macro request, may be 'null'.
	 *
	 * @return the current macro request, may be 'null'.
	 */
	public MacroRequest getRequest() {
		return request;
	}

	/**
	 * Returns the current base directory the Macro is operating against.
	 *
	 * @return the current base directory the Macro is operating against.
	 */
	public File getBasedir() {
		return request == null ? new File(".") : request.getBasedir();
	}

	/**
	 * Returns the file path to use for storing attached includes.
	 *
	 * @return the file path to use for storing attached includes, or 'null' if attaching includes as file is not supported.
	 */
	public File getAttachedIncludesPath() {
		return attachedIncludesPath;
	}

	/**
	 * Resolves the given path to an URL that can be fetched via {@link UrlFetcher}.
	 * <p/>
	 * Note: Under the hood this method delegates to {@link ResourceResolver#findSource(java.io.File, String)}.
	 *
	 * @param path the path to resolve.
	 * @return an URL that can be fetched.
	 */
	public URL resolvePath(String path) {
		return ResourceResolver.findSource(getBasedir(), path);
	}

	/**
	 * Resolves the given local path and returns the corresponding file.
	 * @param localPath the local path to resolve.
	 * @return the resolved file or 'null' if the path couldn't be resolved to a local file or directory.
	 */
	public File resolveLocalPath(String localPath) {
		final URL url = resolvePath(localPath);
		try {
			return "file".equalsIgnoreCase(url.getProtocol()) ? new File(url.toURI()) : null;
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	/**
	 * Fetches the given url and converts the content to a string.
	 *
	 * @param url the url to fetch.
	 * @return url content as string.
	 * @throws java.io.IOException In case of IO failed.
	 */
	public byte[] fetchUrl(URL url) throws IOException {
		final InputStream inputStream = UrlFetcher.getSource(url, false);
		return IOUtil.toByteArray(inputStream);
	}

	/**
	 * Fetches the given url and converts the content to a string.
	 *
	 * @param path the path to fetch.
	 * @return the content as byte array.
	 * @throws java.io.IOException In case of IO failed.
	 */
	public byte[] fetch(String path) throws IOException {
		return fetchUrl(resolvePath(path));
	}

	/**
	 * Fetches the given url and converts the content to a string.
	 *
	 * @param url the url to fetch.
	 * @return url content as string.
	 * @throws java.io.IOException In case of IO failed.
	 */
	public String fetchUrlText(URL url) throws IOException {
		return UrlFetcher.readerToString(UrlFetcher.getReadableSource(url), true);
	}

	/**
	 * Fetches the given path and converts the content to a string.
	 *
	 * @param path the path to fetch.
	 * @return the content as string.
	 * @throws java.io.IOException In case of IO failed.
	 */
	public String fetchText(String path) throws IOException {
		return fetchUrlText(resolvePath(path));
	}

	/**
	 * Returns a set of all resources that were attached to the current sink.
	 *
	 * @return a set of all resources that were attached to the current sink.
	 */
	public Set<String> getAttachedResources() {
		synchronized (attachedResources) {
			Set<String> resources = attachedResources.get(sink);
			if (resources == null) attachedResources.put(sink, resources = new HashSet<String>());
			return resources;
		}
	}

	/**
	 * Attempts to attach the given CSS (cascading style sheet).
	 *
	 * @param fileName   a unique name to reference the content. If the name is empty the css will be inlined.
	 * @param cssContent the content to attach, either in-lined or using the filename reference.
	 *                   If the css content starts with "classpath:", it is interpreted as a classpath URL to include
	 *                   instead of the real content.
	 * @return a status whether attaching the cascading style sheets is supported by the underlying sink.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public boolean attachCss(String fileName, String cssContent) throws IOException {
		final Set<String> resources = getAttachedResources();
		if (resources.contains(fileName)) {
			getLog().debug("Will not attach a CSS of name '" + fileName + "' twice within the same output.");
			return true;
		}

		if (isHTMLCapableSink()) {
			StringBuilder html = new StringBuilder(64).append("<style type=\"text/css\">").append(LB);
			String link = StringUtils.isEmpty(fileName) ? null : attachContent(fileName, cssContent);
			if (link == null) {
				if (cssContent.startsWith("classpath:"))
					cssContent = IOUtil.toString(UrlFetcher.getReadableSource(resolvePath(cssContent)));
				html.append(cssContent);
			} else
				html.append("   @import url('").append(link).append("');");
			html.append(LB).append("</style>").append(LB);

			sink.rawText(html.toString());
			if (!StringUtils.isEmpty(fileName)) resources.add(fileName);
			return true;
		}

		return false;
	}

	/**
	 * Attempts to attach the given client side java script.
	 *
	 * @param fileName      a unique name to reference the script.
	 * @param scriptContent the content to attach, either in-lined or using the filename reference.
	 *                      If the script content starts with "classpath:", it is interpreted as a classpath URL to include
	 *                      instead of the real content.
	 * @return a status whether attaching the script is supported by the underlying sink.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public boolean attachJs(String fileName, String scriptContent) throws IOException {
		return attachScript(fileName, scriptContent, null, null);
	}

	/**
	 * Attempts to attach the given client side script.
	 *
	 * @param fileName       a unique name to reference the script. If the name is empty the script will be inlined.
	 * @param scriptContent  the content to attach, either in-lined or using the filename reference.
	 *                       If the script content starts with "classpath:", it is interpreted as a classpath URL to include
	 *                       instead of the real content.
	 * @param scriptLanguage the script language, defaults to "javascript".
	 * @param contentType    the content type, defaults to "text/javascript".
	 * @return a status whether attaching the script is supported by the underlying sink.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public boolean attachScript(String fileName, String scriptContent, String scriptLanguage, String contentType) throws IOException {
		final Set<String> resources = getAttachedResources();
		if (resources.contains(fileName)) {
			getLog().debug("Will not attach a Script of name '" + fileName + "' twice within the same output.");
			return true;
		}

		if (isHTMLCapableSink()) {
			if (scriptLanguage == null) scriptLanguage = "javascript";
			if (contentType == null) contentType = "text/javascript";

			StringBuilder html = new StringBuilder(64).append("<script " +
					"type=\"").append(contentType).append("\" " +
					"language=\"").append(scriptLanguage).append('"');

			String link = StringUtils.isEmpty(fileName) ? null : attachContent(fileName, scriptContent);
			if (link == null) {
				if (scriptContent.startsWith("classpath:"))
					scriptContent = IOUtil.toString(UrlFetcher.getReadableSource(resolvePath(scriptContent)));
				html.append('>').append(LB).append(scriptContent).append(LB);
			} else
				html.append(" src=\"").append(link).append("\">");
			html.append("</script>").append(LB);

			sink.rawText(html.toString());
			if (!StringUtils.isEmpty(fileName)) resources.add(fileName);
			return true;
		}

		return false;
	}


	/**
	 * Attempts to attach the given content as a file and returns the relative link to the content on success.
	 * <p/>
	 * Content can only be attached when building sites, as there has to be a resolvable resource directory
	 * that is used to put the attached content.
	 * <p/>
	 * Files of the same name are overwritten if they already exist in the output directory.
	 * <p/>
	 * When content is stored, it is saved in 2 versions, uncompressed and gzipped. The latter is provided to reduce
	 * transfer sizes on supporting web-servers and is created if the compressed size is smaller than 95% of the uncompressed size.
	 *
	 * @param fileName a unique name to reference the content.
	 * @param content  the content to attach.
	 *                 If the content starts with "classpath:", it is interpreted as a classpath URL to include
	 *                 instead of the real content.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachContent(String fileName, String content) throws IOException {
		if (attachedIncludesPath == null) return null;
		if (StringUtils.isEmpty(content)) throw new IllegalArgumentException("Cannot attach 'null'. No content was specified.");

		if (content.startsWith("classpath:"))
			return attachBinaryContent(fileName, content);

		final Log log = getLog();
		Charset charset = getCharset(), utf8 = Charset.forName("UTF-8");

		// Handle charset for XML, it always defaults to UTF-8, overriding any user supplied charset.
		if (content.trim().startsWith("<")) {
			Matcher matcher = compile("\\<\\?xml.*encoding=['\"]{1}(.+)['\"]{1}").matcher(content);
			if (matcher.find())
				charset = Charset.forName(matcher.group(1));
			else if (content.contains("/>") || !content.contains("<html"))
				charset = utf8;
		}

		CharsetEncoder charsetEncoder = getCharset().newEncoder();
		if (!charsetEncoder.canEncode(content)) {
			log.warn("The selected charset of '" + charset.name() + "' cannot be used to encode the specified content. " +
					"Will change the charset to '" + utf8 + '\'');
			charset = utf8;
			charsetEncoder = charset.newEncoder();
		}

		if (log.isDebugEnabled()) log.debug("Selected charset for '" + fileName + "' using '" + charset + '\'');
		try {
			ByteBuffer buffer = charsetEncoder.encode(CharBuffer.wrap(content));
			final ByteArrayInputStream stream = new ByteArrayInputStream(buffer.array(), buffer.arrayOffset(), buffer.remaining());
			return attachBinaryContent(fileName, stream);
		} catch (RuntimeException e) {
			if (log.isDebugEnabled()) log.debug("Content that will not be stored:" + LB + content);
			throw e;
		}
	}

	/**
	 * Overloads {@link #attachBinaryContent(URL)}.
	 *
	 * @param url the url or path to fetch the binary content from.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(String url) throws IOException {
		return attachBinaryContent(resolvePath(url));
	}

	/**
	 * Overloads {@link #attachBinaryContent(String, URL)}.
	 *
	 * @param fileName a unique name to reference the content.
	 * @param url      the url or path to fetch the binary content from.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(String fileName, String url) throws IOException {
		return attachBinaryContent(fileName, resolvePath(url));
	}

	/**
	 * Overloads {@link #attachBinaryContent(String, URL)}.
	 *
	 * @param sourceUrl an URL to fetch the binary content that should be attached.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(URL sourceUrl) throws IOException {
		return attachBinaryContent(sourceUrl.getPath(), sourceUrl);
	}

	/**
	 * Overloads {@link #attachBinaryContent(String, java.io.InputStream)}.
	 *
	 * @param fileName  a unique name to reference the content.
	 * @param sourceUrl an URL to fetch the binary content that should be attached.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(String fileName, URL sourceUrl) throws IOException {
		return attachBinaryContent(fileName, UrlFetcher.getSource(sourceUrl, false));
	}

	/**
	 * Overloads {@link #attachBinaryContent(String, java.io.InputStream)}.
	 *
	 * @param fileName      a unique name to reference the content.
	 * @param binaryContent the content to attach.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(String fileName, byte[] binaryContent) throws IOException {
		return attachBinaryContent(fileName, new ByteArrayInputStream(binaryContent));
	}

	/**
	 * Attempts to attach the given content as a file and returns the relative link to the content on success.
	 * <p/>
	 * Content can only be attached when building sites, as there has to be a resolvable resource directory
	 * that is used to put the attached content.
	 * <p/>
	 * Files of the same name are overwritten if they already exist in the output directory.
	 * <p/>
	 * When content is stored, it is saved in 2 versions, uncompressed and gzipped. The latter is provided to reduce
	 * transfer sizes on supporting web-servers and is created if the compressed size is smaller than 95% of the uncompressed size.
	 *
	 * @param fileName      a unique name to reference the content.
	 * @param binaryContent the content to attach.
	 * @return A relative link to the content or 'null' if attaching content is not supported by the underlying environment.
	 * @throws java.io.IOException In case of disk IO failed.
	 */
	public String attachBinaryContent(String fileName, InputStream binaryContent) throws IOException {
		if (attachedIncludesPath == null) return null;
		if (binaryContent == null) throw new IllegalArgumentException("Cannot attach 'null'. Binary content is not specified.");
		if (StringUtils.isEmpty(fileName)) {
			throw new IllegalArgumentException("Cannot attach binary content '" + binaryContent + "' when filename is empty.");
		}

		final Log log = getLog();
		final boolean debug = log.isDebugEnabled();

		final File path = attachedIncludesPath.getCanonicalFile();
		final File targetFile;
		try {
			targetFile = createTargetFile(path, fileName);
		} catch (IOException e) {
			IOException ioException = new IOException(
					"Failed building target path to attach file " + fileName + " under base path " + path);
			ioException.initCause(e);
			throw ioException;
		}

		if (debug) log.debug("About to store content to '" + targetFile + '\'');
		final File compressedTargetFile = new File(targetFile + ".gz");
		final OutputStream outputStream = new FileOutputStream(targetFile);
		try {
			final OutputStream compressedOutputStream = new GZIPOutputStream(new FileOutputStream(compressedTargetFile));
			try {
				int read = 0;
				byte[] buffer = new byte[64 * 1024];
				while ((read = binaryContent.read(buffer)) != -1) {
					outputStream.write(buffer, 0, read);
					compressedOutputStream.write(buffer, 0, read);
				}
			} finally {
				compressedOutputStream.close();
			}
		} finally {
			outputStream.close();
		}

		final double compressionRatio = (double) compressedTargetFile.length() / (double) targetFile.length();
		if (compressionRatio > 0.95) {
			if (debug) {
				log.debug("The size of '" + compressedTargetFile + "' is " + Math.round(100 * compressionRatio) +
						"% the size of the uncompressed file, removing the compressed version as it is no improvement.");
			}
			if (!compressedTargetFile.delete()) compressedTargetFile.deleteOnExit();
		} else {
			if (debug) {
				log.debug("Generated a compressed version of the attached content '" + compressedTargetFile + "' having " +
						Math.round(100 * compressionRatio) + "% the size of the uncompressed file.");
			}
		}

		getAttachedResources().add(fileName);

		final URI baseUri = createBaseUri(path);
		return baseUri.relativize(targetFile.toURI()).toASCIIString();
	}

	private static URI createBaseUri(File path) {
		final String attachedIncludesDirectory = getAttachedIncludesDirectory();
		if (!StringUtils.isEmpty(attachedIncludesDirectory)) {
			String pathString = path.getAbsolutePath().replace('\\', '/');
			if (pathString.endsWith(attachedIncludesDirectory))
				pathString = pathString.substring(0, pathString.length() - attachedIncludesDirectory.length());
			return new File(pathString).toURI();
		} else
			return path.toURI();
	}

	static File createTargetFile(File path, String fileName) throws IOException {
		final Log log = getLog();
		// Detect URIs and extract the path portion
		if (fileName.matches("^[a-zA-Z]{2,}:[/]{1,3}.+")) {
			if (log.isDebugEnabled()) {
				log.debug("The given filename '" + fileName + "' seems to be an URI, trying to extract the URI path for use as " +
						"filename instead of the complete URI.");
			}
			fileName = URI.create(fileName).getPath();
			if (fileName.contains("!/")) fileName = fileName.substring(fileName.indexOf("!/") + 2);
		}

		// Convert absolute to relative paths (linux and windows.. Note: Must be that order!)
		if (fileName.startsWith("/")) fileName = fileName.substring(1);
		if (fileName.matches("^[a-zA-Z]{1}:.+")) fileName = fileName.substring(2);

		final File targetFile = new File(path, fileName).getCanonicalFile(), targetDirectory = targetFile.getParentFile();
		if (!targetFile.getAbsolutePath().startsWith(path.getAbsolutePath()))
			throw new IllegalArgumentException("Cannot store target file " + targetFile + " as it is not located below " + path);
		if (targetDirectory == null || (!targetDirectory.isDirectory() && !targetDirectory.mkdirs())) {
			log.warn("Cannot create target directory '" + targetDirectory + "' to store content for file '" + fileName + '\'');
			throw new IllegalArgumentException("Cannot store target file " + targetFile + " as the parent " +
					"directory does not exist and cannot be created. Parent directory is " + targetDirectory);
		}
		return targetFile;
	}

	@Override
	public String toString() {
		return "Globals{" +
				"charset=" + charset +
				", charsetIsUserSupplied=" + charsetIsUserSupplied +
				", request=" + request +
				", attachedIncludesPath=" + attachedIncludesPath +
				", sink=" + sink +
				", logger=" + logger +
				'}';
	}
}
