package org.tinyjee.maven.dim.spi.backport;

import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.net.URL;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.atomic.AtomicReference;

/**
 * Minimum efforts backport of ServiceLoader found in Java 1.6 to be used in Java 1.5.
 * <p/>
 * (May not behave in the same way and is meant to support adapter bindings only)
 * <p/>
 * Note: This implementation searches inside "META-INF/services/typeName" and "typePackage/typeName.services".
 * The latter is not standards compliant, but it works on Android where getting something into META-INF is
 * quite an issue.
 *
 * @author Juergen_Kellerer, 2011-09-27
 */
public class ServiceLoader {

	private ServiceLoader() {
	}

	/**
	 * Loads the services that are specified by the given service type.
	 *
	 * @param serviceType the service type to load.
	 * @param <T>         the expected java type.
	 * @return a collection of service implementations.
	 */
	public static <T> Iterable<T> load(Class<T> serviceType) {
		return new LazyIterable<T>(serviceType);
	}

	private static final class LazyIterable<T> implements Iterable<T> {

		private static final String PREFIX = "META-INF/services/";

		private final Class<T> serviceType;
		private final AtomicReference<List<T>> implementations = new AtomicReference<List<T>>();

		private LazyIterable(Class<T> serviceType) {
			this.serviceType = serviceType;
		}

		public Iterator<T> iterator() {
			if (implementations.get() == null) {
				try {
					// we perform eager initialization of all instances to simplify the implementation.
					List<T> instances = new ArrayList<T>();
					String[] resourceNames = {PREFIX + serviceType.getName(), serviceType.getName().replace('.', '/') + ".services"};
					ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
					if (classLoader == null)
						classLoader = serviceType.getClassLoader();

					for (String resourceName : resourceNames)
						for (Enumeration<URL> urls = classLoader.getResources(resourceName); urls.hasMoreElements(); ) {
							BufferedReader reader = new BufferedReader(new InputStreamReader(urls.nextElement().openStream(), "utf-8"));
							try {
								String implementation;
								while ((implementation = reader.readLine()) != null) {
									int commentIndex = (implementation = implementation.trim()).indexOf('#');
									if (commentIndex != -1)
										implementation = implementation.substring(0, commentIndex).trim();
									if (implementation.length() != 0)
										instances.add(serviceType.cast(classLoader.loadClass(implementation).newInstance()));
								}
							} finally {
								reader.close();
							}
						}

					implementations.compareAndSet(null, instances);
				} catch (Exception e) {
					throw new RuntimeException(e);
				}
			}

			return implementations.get().iterator();
		}
	}
}
