package org.tinyjee.maven.dim.utils;

import com.thoughtworks.qdox.JavaDocBuilder;
import com.thoughtworks.qdox.model.JavaClass;
import com.thoughtworks.qdox.model.JavaPackage;
import com.thoughtworks.qdox.model.JavaSource;
import org.apache.maven.doxia.Doxia;
import org.tinyjee.maven.dim.spi.Globals;
import org.tinyjee.maven.dim.spi.ResourceResolver;

import java.io.File;
import java.io.IOException;
import java.io.Reader;
import java.io.UnsupportedEncodingException;
import java.lang.ref.SoftReference;
import java.net.URISyntaxException;
import java.net.URL;
import java.net.URLDecoder;
import java.util.*;

import static java.util.Arrays.asList;

/**
 * Is a singleton that loads java classes at source level using qdox.
 *
 * @author Juergen_Kellerer, 2011-10-11
 */
public class JavaClassLoader {

	private static final JavaClassLoader instance = new JavaClassLoader();

	public static JavaClassLoader getInstance() {
		return instance;
	}

	private SoftReference<SharedBuildHolder> sharedBuilderReference = new SoftReference<SharedBuildHolder>(null);
	private final Set<String> sourceUrls = new LinkedHashSet<String>();
	private final Set<String> sourceTrees = new LinkedHashSet<String>();

	private JavaClassLoader() {
	}

	public synchronized JavaSource addSource(File basePath, String pathOrUrl) throws IOException {
		return addSource(basePath, pathOrUrl, false);
	}

	public synchronized JavaSource addSource(File basePath, String pathOrUrl, boolean doNotLoadSourceTree) throws IOException {
		return addSource(ResourceResolver.findSource(basePath, pathOrUrl), doNotLoadSourceTree);
	}

	public synchronized JavaSource addSource(URL sourceUrl, Reader sourceContent) {
		String sourceInfo = sourceUrl.toExternalForm();
		sourceUrls.add(sourceInfo);
		return getSharedBuilder(false).addSource(sourceContent, sourceInfo);
	}

	public synchronized JavaSource addSource(URL sourceUrl, boolean doNotLoadSourceTree) throws IOException {
		JavaSource javaSource = getSource(sourceUrl);
		if (javaSource != null)
			return javaSource;

		try {
			if (!doNotLoadSourceTree && "file".equals(sourceUrl.getProtocol())) {
				// The builder does not de-duplicate already added source, thus we must use a fresh builder here
				javaSource = createNewBuilder(true).addSource(sourceUrl);

				final String packagePath = javaSource.getPackageName().replace('.', '/');
				final String url = sourceUrl.toString(), rootUrl = "".equals(packagePath) ?
						sourceUrl.toURI().resolve(".").toASCIIString() :
						url.substring(0, url.indexOf(packagePath));
				final File sourceFolder = new File(new URL(rootUrl).toURI());
				addSourceTree(sourceFolder);

				javaSource = getSource(sourceUrl);
				if (javaSource == null) {
					Globals.getLog().warn("Failed getting java source " + sourceUrl + " after adding source path " + sourceFolder +
							", will load the source directly which may lead to duplicates when selecting JavaClasses later.");
				}
			}

			if (javaSource == null) {
				sourceUrls.add(sourceUrl.toString());
				javaSource = getSharedBuilder(false).addSource(sourceUrl);
			}

			return javaSource;
		} catch (URISyntaxException e) {
			throw new RuntimeException(e);
		}
	}

	public synchronized void addSourceTree(File sourceFolder) {
		if (sourceTrees.add(sourceFolder.toString()))
			getSharedBuilder(false).addSourceTree(sourceFolder);
	}

	public JavaSource getSource(URL sourceUrl) {
		try {
			final String urlCharset = Globals.getInstance().getCharset().name();
			final String urlString = sourceUrl.toExternalForm(), decodedUrlString = URLDecoder.decode(urlString, urlCharset);
			for (JavaSource javaSource : getSources()) {
				URL otherUrl = javaSource.getURL();
				String otherUrlString = otherUrl == null ? "" : otherUrl.toExternalForm();
				if (otherUrlString.equals(urlString) || otherUrlString.equals(decodedUrlString))
					return javaSource;
			}
		} catch (UnsupportedEncodingException e) {
			throw new RuntimeException(e);
		}
		return null;
	}

	public synchronized JavaSource[] getSources() {
		return getSharedBuilder(true).getSources();
	}

	public SortedSet<JavaClass> getClasses() {
		final JavaClass[] classes;
		synchronized (this) {
			classes = getSharedBuilder(true).getClasses();
		}

		return new TreeSet<JavaClass>(asList(classes));
	}

	public SortedSet<JavaPackage> getPackages() {
		final JavaPackage[] packages;
		synchronized (this) {
			packages = getSharedBuilder(true).getPackages();
		}

		final TreeSet<JavaPackage> javaPackages = new TreeSet<JavaPackage>(new Comparator<JavaPackage>() {
			public int compare(JavaPackage o1, JavaPackage o2) {
				return o1.getName().compareTo(o2.getName());
			}
		});
		javaPackages.addAll(asList(packages));
		return javaPackages;
	}

	public synchronized void clear() {
		sourceTrees.clear();
		sourceUrls.clear();
		sharedBuilderReference = new SoftReference<SharedBuildHolder>(null);
	}

	JavaDocBuilder getSharedBuilder(boolean needsImmediateUpToDateCheck) {
		JavaDocBuilder builder;
		SharedBuildHolder holder = sharedBuilderReference.get();

		if (!needsImmediateUpToDateCheck)
			scheduleUpToDateCheck();
		else if (holder != null && !holder.isBuilderUpToDate())
			holder = null;

		if (holder == null) {
			builder = createNewBuilder(false);
			sharedBuilderReference = new SoftReference<SharedBuildHolder>(new SharedBuildHolder(builder));
		} else
			builder = holder.getBuilder();

		return builder;
	}

	void scheduleUpToDateCheck() {
		final SharedBuildHolder holder = sharedBuilderReference.get();
		if (holder != null) holder.resetLastUpToDateCheck();
	}

	synchronized JavaDocBuilder createNewBuilder(boolean forceEmpty) {
		JavaDocBuilder javaDocBuilder = new JavaDocBuilder();
		javaDocBuilder.setEncoding(Globals.getInstance().getCharset().name());

		if (!forceEmpty) {
			try {
				for (String sourceUrl : sourceUrls)
					javaDocBuilder.addSource(new URL(sourceUrl));
				for (String sourceTree : sourceTrees)
					javaDocBuilder.addSourceTree(new File(sourceTree));
			} catch (IOException e) {
				throw new RuntimeException(e);
			}
		}

		return javaDocBuilder;
	}

	/**
	 * Holds a shared builder with the option to do a file sync. check on retrieve.
	 */
	private static class SharedBuildHolder {

		private static final int MS_BETWEEN_CHECKS = 1000;

		private final JavaDocBuilder builder;

		private long lastUpToDateCheck;
		private final Map<String, Long> sourceModifiedTimes = new HashMap<String, Long>();

		private SharedBuildHolder(JavaDocBuilder builder) {
			if (builder == null) throw new IllegalArgumentException("The wrapped builder cannot be set to 'null'");
			this.builder = builder;

			isBuilderUpToDate();
		}

		JavaDocBuilder getBuilder() {
			return builder;
		}

		synchronized void resetLastUpToDateCheck() {
			lastUpToDateCheck = 0;
		}

		synchronized boolean isBuilderUpToDate() {
			final long time = System.currentTimeMillis();
			if (time - lastUpToDateCheck > MS_BETWEEN_CHECKS) {
				final JavaSource[] sources = builder.getSources();
				for (JavaSource javaSource : sources) {
					final URL url = javaSource.getURL();
					try {
						if (url != null && "file".equalsIgnoreCase(url.getProtocol())) {
							final String key = url.toString();
							final long sourceLastModified = url.openConnection().getLastModified();
							final Long capturedLastModified = sourceModifiedTimes.get(key);

							if (capturedLastModified == null)
								sourceModifiedTimes.put(key, sourceLastModified);
							else if (capturedLastModified != sourceLastModified)
								return false;
						}
					} catch (IOException e) {
						Globals.getLog().warn("Failed to get last modified date for URL " + url, e);
					}
				}
				lastUpToDateCheck = time;
			}
			return true;
		}
	}
}
