package org.tinyjee.maven.dim.utils;

import org.apache.maven.doxia.logging.Log;
import org.tinyjee.maven.dim.spi.Globals;

import java.util.LinkedHashMap;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * Extends LinkedHashMap with the ability to print the map's contents as formatted string.
 * <p/>
 * Contents are automatically printed to the build log when a non-existing key is queried or
 * when {@link #printContent()} is called.
 *
 * @author Juergen_Kellerer, 2011-09-30
 */
public class PrintableMap<K, V> extends LinkedHashMap<K, V> {

	private static final long serialVersionUID = 6815282986526871197L;
	private static final String NEWLINE = System.getProperty("line.separator", "\n");

	private final String name;

	public PrintableMap(String name) {
		this.name = name;
	}

	public PrintableMap(Map<? extends K, ? extends V> map, String name) {
		super(map);
		this.name = name;
	}

	@Override
	public V get(Object key) {
		V value = super.get(key);

		if (value == null && !containsKey(key)) {
			String mapContent = getContentAsString();
			final Log logger = Globals.getLog();
			logger.warn("Didn't find the requested key '" + key + "' inside this '" + name + "' map." + NEWLINE +
					"Contained values are: " + mapContent);
		}

		return value;
	}

	public void printContent() {
		Globals.getLog().info(getContentAsString());
	}

	public String getContentAsString() {
		String mapContent = toString();
		StringTokenizer tokenizer = new StringTokenizer(mapContent, "{},", true);

		int indentDepth = 0;
		StringBuilder buffer = new StringBuilder(mapContent.length());
		while (tokenizer.hasMoreTokens()) {
			String token = tokenizer.nextToken();
			switch (token.charAt(0)) {
				case '{':
					buffer.append(token);
					indentDepth++;
					break;
				case ',':
					buffer.append(token);
					break;
				case '}':
					indentDepth--;
				default:
					buffer.append(NEWLINE);
					for (int i = 0; i < indentDepth; i++) buffer.append("   ");
					buffer.append(token.trim());
			}
		}
		mapContent = buffer.toString();
		return mapContent;
	}
}
