package org.tinyjee.maven.dim.utils;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import javax.xml.namespace.NamespaceContext;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import java.io.StringWriter;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

/**
 * Implements {@link XPathEvaluator}.
 *
 * @author Juergen_Kellerer, 2011-10-12
 */
public class XPathEvaluatorImplementation implements XPathEvaluator {

	public static final String DEFAULT_PREFIX = "default";

	public static String serializeNode(Node node, boolean omitDeclaration, boolean standalone) {
		try {
			StringWriter buffer = new StringWriter(4096);
			Transformer transformer = TransformerFactory.newInstance().newTransformer();
			if (omitDeclaration) transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");
			if (standalone) transformer.setOutputProperty(OutputKeys.STANDALONE, "yes");
			transformer.transform(new DOMSource(node), new StreamResult(buffer));
			return buffer.toString();
		} catch (TransformerException e) {
			throw new RuntimeException(e);
		}
	}

	private final Node context;
	private final XPath xPath;

	/**
	 * Creates a new evaluator around the given document.
	 *
	 * @param document the document to evaluate on.
	 */
	public XPathEvaluatorImplementation(final Document document) {
		context = document;
		xPath = XPathFactory.newInstance().newXPath();
		xPath.setNamespaceContext(new NamespaceContext() {
			public String getNamespaceURI(String prefix) {
				return DEFAULT_PREFIX.equals(prefix) ? document.lookupNamespaceURI(null) : document.lookupNamespaceURI(prefix);
			}

			public String getPrefix(String namespaceURI) {
				return namespaceURI.equals(document.lookupNamespaceURI(null)) ? DEFAULT_PREFIX : document.lookupPrefix(namespaceURI);
			}

			public Iterator getPrefixes(String namespaceURI) {
				return Collections.singleton(getPrefix(namespaceURI)).iterator();
			}
		});
	}

	private XPathEvaluatorImplementation(XPath xPath, Node context) {
		this.xPath = xPath;
		this.context = context;
	}

	public String findText(String xpathExpression) {
		try {
			return (String) xPath.evaluate(xpathExpression, context, XPathConstants.STRING);
		} catch (XPathExpressionException e) {
			throw new RuntimeException(e);
		}
	}

	public List<Node> findNodes(String xpathExpression) {
		try {
			final NodeList nodeList = (NodeList) xPath.evaluate(xpathExpression, context, XPathConstants.NODESET);
			return asList(nodeList);
		} catch (XPathExpressionException e) {
			throw new RuntimeException(e);
		}
	}

	public String serialize(Node node) {
		return serializeNode(node, true, true);
	}

	public List<Node> asList(final NodeList nodeList) {
		return new NodeListAdapter(nodeList);
	}

	public XPathEvaluator newEvaluator(Node node) {
		return new XPathEvaluatorImplementation(xPath, node);
	}
}
